﻿using DbCommon.BusinessCore.BaseCore;
using ProjectManagementSystem.Common.Config;
using ProjectManagementSystem.Common.Extenions;
using ProjectManagementSystem.Common.Function;
using ProjectManagementSystem.Common.Logger;
using ProjectManagementSystem.Common.Models;
using ProjectManagementSystem.Common.Models.Crms;
using ProjectManagementSystem.Common.Service;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace ProjectManagementSystem.Common.Core
{

    public sealed class PackInteractionCrms8952A : PackInteractionCrmsBase
    {
        public override string GetNextLocation(PackStationV2 data, out string templateName, out string msg)
        {
            var config = data.Config;
            templateName = null;

            var result = ReadPlcValue(config, "PLC一键驶离", out msg);

            DeviceControl.Instance.Communication.InfoVcuReleaseDictionary.TryGetValue(data.AgvAtStationAgvId, out var infoVcuRelease);
            if (infoVcuRelease == null)
            {
                msg = $"{msg} 获取AGV{data.AgvAtStationAgvId}放行失败";
            }
            else
            {
                if (infoVcuRelease.DataValid)
                {
                    msg = $"{msg} 获取AGV{data.AgvAtStationAgvId}放行成功";
                }
                else
                {
                    msg = $"{msg} 获取AGV{data.AgvAtStationAgvId}放行超时";
                }
            }

            var vcuRelease = infoVcuRelease != null && infoVcuRelease.DataValid;
            if (result > 0 || vcuRelease)
            {
                if (config.RelationStationID2 <= 0)
                {
                    msg = $"{msg} 一键驶离未配置关联点2";
                    return null;
                }

                templateName = "一键驶离";
                return data.Id;
            }

            //没有一键驶离信号

            var nextLocation = base.GetNextLocation(data, out templateName, out msg);
            if (string.IsNullOrWhiteSpace(nextLocation))
            {
                return null;
            }

            ReadValue(config, "PLC物料", string.Empty, out var plcMaterial, out var msg2);
            if (string.IsNullOrWhiteSpace(plcMaterial))
            {
                msg = $"{msg} {msg2}";
                return null;
            }

            var resBind = Crms.PmsTaskService.BindAgvMaterial(new AgvMaterialDto() { AgvId = data.AgvAtStationAgvId, MaterialId = "默认货物", MaterialBarcode = plcMaterial });
            if (resBind.code != 20000)
            {
                msg = $"{msg} {msg2} 放行绑定物料失败 {resBind?.message}";
                return null;
            }

            msg = $"{msg} {msg2} 放行绑定物料 AGV{data.AgvAtStationAgvId} {plcMaterial}";

            return nextLocation;
        }
        public override bool StopAgvEntering(PackStationV2 data, out string msg)
        {
            var result = ReadPlcValue(data.Config, "PLC急停", out msg);
            if (result == failedCode || result == 0)
            {
                return true;
            }

            return false;
        }
        public override bool StopAgvLeaving(PackStationV2 data, out string msg)
        {
            var result = ReadPlcValue(data.Config, "PLC急停", out msg);
            if (result == failedCode || result == 0)
            {
                return true;
            }

            return false;
        }

        public override void Entering(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            if (agv.leaver == 1) return;

            //一键驶离的AGV经过的自动工位不停
            var config = data.Config;
            var result = taskList.Exists(d => d.Carrier == agv.carrier && d.TaskBook == "一键驶离");
            var allowStatus = result ? 1 : 0;

            if (agv.leaver != allowStatus)
            {
                var result2 = Crms.PmsApi.AllowAgv(data.Id, agv.carrier, 0, allowStatus, 0);
                var msg = $"一键驶离任务{result} {GetAgvDataString(agv.AgvData)} 工位点{GetAllowStatusString(allowStatus)}{result2.ToChineseString()}";
                CLog.Instance.TaskLog.WriteInfo($"{config.LocationCode} {msg}");
            }
        }
        public override void WaitStation(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            if (agv.leaver == 1) return;

            var config = data.Config;
            var result = taskList.Exists(d => d.Carrier == agv.carrier && d.TaskBook == "一键驶离");
            var allowStatus = result ? 1 : 0;
            var msg = $"一键驶离任务{result} {GetAgvDataString(agv.AgvData)}";

            if (agv.leaver != allowStatus)
            {
                var result2 = Crms.PmsApi.AllowAgv(data.Id, agv.carrier, 0, allowStatus, 0);
                msg = $"{msg} 工位点{GetAllowStatusString(allowStatus)}{result2.ToChineseString()}";
            }

            base.WaitStation(data, agv, taskList);
        }
    }
    public sealed class PackInteractionCrms8952S : PackInteractionCrmsBase
    {
        private PmsTaskInfoManager pmsTaskInfoManger = new PmsTaskInfoManager();
        public override void Heartbeat(PackStationV2 data, bool agvHeartbeat)
        {
        }
        public override void CustomProcess(PackStationV2 data, bool firstStart, bool[] forceWrite)
        {
            //进入工位区域的所有AGV
            var agvAtStationDataList = data.AgvList?
                .Where(d => d.agvPositionStatus == EPackAgvPositionStatus.Entering
                || d.agvPositionStatus == EPackAgvPositionStatus.StationVertex
                || d.agvPositionStatus == EPackAgvPositionStatus.Leaving
                || d.agvPositionStatus == EPackAgvPositionStatus.ReleaseVertex)?
                .ToList();

            if (agvAtStationDataList?.Count > 0)
            {
                var config = data.Config;

                var pmsTaskList = pmsTaskInfoManger.GetList(d => d.LocationCode == data.Id);
                foreach (var agv in agvAtStationDataList)
                {
                    var plcRealese = pmsTaskList.Exists(d => d.Carrier == agv.carrier);

                    if (!plcRealese)
                    {
                        var result = ReadPlcValue(config, "PLC放行", out var msg);
                        if (result > 0)
                        {
                            ReadValue(config, "PLC物料", string.Empty, out var plcMaterial, out var msg2);
                            if (!string.IsNullOrWhiteSpace(plcMaterial))
                            {
                                //缓行过程中记录PLC放行
                                var resBind = Crms.PmsTaskService.BindAgvMaterial(new AgvMaterialDto() { AgvId = agv.carrier, MaterialId = "默认货物", MaterialBarcode = plcMaterial });
                                var result2 = resBind?.code == 20000;
                                CLog.Instance.TaskLog.WriteInfo($"{config.LocationCode} 放行绑码{result2.ToChineseString()} AGV{agv.carrier} {plcMaterial}");

                                if (result2)
                                {
                                    var dto = new PmsTaskInfoDto();
                                    dto.LocationCode = data.Id;
                                    dto.Carrier = agv.carrier;
                                    dto.MaterialId = plcMaterial;
                                    Crms.PmsTaskService.TaskAppointment(dto);
                                }
                            }
                        }
                    }
                }

                //已经放行的AGV不算到位
                var releaseAgvIds = pmsTaskList.Select(d => d.Carrier).ToArray();
                data.AgvAtStationData = agvAtStationDataList.FirstOrDefault(d => !releaseAgvIds.Contains(d.carrier));
            }
            else
            {
                data.AgvAtStationData = null;
            }

            base.CustomProcess(data, firstStart, forceWrite);
        }
        public override string GetNextLocation(PackStationV2 data, out string templateName, out string msg)
        {
            var config = data.Config;
            templateName = null;

            msg = null;
            string nextLocation = config.LocationCode.GetLocationMember("下个工位");
            if (string.IsNullOrWhiteSpace(nextLocation))
            {
                msg = $"{msg} 没配置下个工位";
                return null;
            }

            templateName = config.LocationCode.GetLocationMember("后任务模板");
            return nextLocation.ToValueArray<string>()[0];
        }
        public override bool StopAgvEntering(PackStationV2 data, out string msg)
        {
            var result = ReadPlcValue(data.Config, "PLC急停", out msg);
            if (result == failedCode || result == 0)
            {
                return true;
            }

            return false;
        }
        public override bool StopAgvLeaving(PackStationV2 data, out string msg)
        {
            var result = ReadPlcValue(data.Config, "PLC急停", out msg);
            if (result == failedCode || result == 0)
            {
                return true;
            }

            return false;
        }

        public override void WaitTask(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
        }
        public override void WaitEnter(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            WaitRequest(data, agv, taskList);
        }
        public override void WaitRequest(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            if (agv.enter == 1) return;

            var config = data.Config;
            var result = ReadPlcValue(config, "PLC允许进入", out var msg);
            var allowStatus = result > 0 ? 1 : 0;
            msg = $"{msg} {GetAgvDataString(agv.AgvData)}";

            if (agv.enter != allowStatus)
            {
                var result2 = Crms.PmsApi.AllowAgv(data.Id, agv.carrier, allowStatus, allowStatus, 0);
                msg = $"{msg} 请求点+工位点{GetAllowStatusString(allowStatus)}{result2.ToChineseString()}";
            }

            if (data.MessageWaitRequest != msg)
            {
                CLog.Instance.TaskLog.WriteInfo($"{config.LocationCode} {msg}");
            }
            data.MessageWaitRequest = msg;
        }
        public override void WaitStation(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            if (agv.leaver == 1) return;

            var config = data.Config;
            var result2 = Crms.PmsApi.AllowAgv(data.Id, agv.carrier, 0, 1, 0);
            CLog.Instance.TaskLog.WriteInfo($"{config.LocationCode} 缓行工位点直接通行 {GetAgvDataString(agv.AgvData)} 工位点通行{result2.ToChineseString()}");

            data.MessageWaitStation = null;
        }
        public override void Leaving(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            if (agv.schedule > 0.99)
            {
                WaitLeave(data, agv, taskList);
            }
        }
        public override void WaitLeave(PackStationV2 data, PackCarrier agv, List<TaskData> taskList)
        {
            if (agv.release == 1) return;

            var config = data.Config;
            var pmsTaskList = pmsTaskInfoManger.GetList(d => d.LocationCode == data.Id && d.Carrier == agv.carrier);
            var plcRealese = pmsTaskList.Count > 0;

            var result = ReadPlcValue(config, "PLC拉绳", out var msg);
            var allowStatus = result == 0 && plcRealese ? 1 : 0;
            msg = $"{msg} PLC放行记录{plcRealese} {GetAgvDataString(agv.AgvData)}";
            var result2 = false;

            if (agv.release != allowStatus)
            {
                result2 = Crms.PmsApi.AllowAgv(data.Id, agv.carrier, 0, 0, allowStatus);
                msg = $"{msg} 释放点{GetAllowStatusString(allowStatus)}{result2.ToChineseString()}";
            }

            if (data.MessageWaitLeave != msg)
            {
                CLog.Instance.TaskLog.WriteInfo($"{config.LocationCode} {msg}");
            }
            data.MessageWaitLeave = msg;


            if (allowStatus == 1 && result2)
            {
                foreach (var item in pmsTaskList)
                {
                    var deleteResult = pmsTaskInfoManger.Delete(item.TaskID);
                    CLog.Instance.TaskLog.WriteInfo($"{config.LocationCode} 删除预约任务{deleteResult.ToChineseString()} {item.TaskID}");
                }
            }
        }
    }
}
